@@ -3,9 +3,10 @@ import asyncio
 import logging
 import pathlib
 import shutil
-from typing import Callable, Dict, Optional
+from typing import Callable, List, Optional
 
 import aiocron
+import aiohttp
 from pyunifiprotect import ProtectApiClient
 from pyunifiprotect.data.nvr import Event
 from pyunifiprotect.data.types import EventType, ModelType
@@ -14,6 +15,27 @@ from pyunifiprotect.data.websocket import WSAction, WSSubscriptionMessage
 logger = logging.getLogger(__name__)
 
 
+class RcloneException(Exception):
+    """Exception class for when rclone does not exit with `0`."""
+
+    def __init__(self, stdout, stderr, returncode):
+        """Exception class for when rclone does not exit with `0`.
+
+        Args:
+          stdout (str): What rclone output to stdout
+          stderr (str): What rclone output to stderr
+          returncode (str): The return code of the rclone process
+        """
+        super().__init__()
+        self.stdout: str = stdout
+        self.stderr: str = stderr
+        self.returncode: int = returncode
+
+    def __str__(self):
+        """Turns excpetion into a human readable form."""
+        return f"Return Code: {self.returncode}\nStdout:\n{self.stdout}\nStderr:\n{self.stderr}"
+
+
 def add_logging_level(levelName: str, levelNum: int, methodName: Optional[str] = None) -> None:
     """Comprehensively adds a new logging level to the `logging` module and the currently configured logging class.
 
@@ -146,10 +168,10 @@ class UnifiProtectBackup:
         retention (str): How long should event clips be backed up for. Format as per the
                          `--max-age` argument of `rclone`
                          (https://rclone.org/filtering/#max-age-don-t-transfer-any-file-older-than-this)
+        ignore_cameras (List[str]): List of camera IDs for which to not backup events
         verbose (int): How verbose to setup logging, see :func:`setup_logging` for details.
         _download_queue (asyncio.Queue): Queue of events that need to be backed up
         _unsub (Callable): Unsubscribe from the websocket callback
-        _camera_names (Dict[str, str]): A map of camera IDs -> camera names
     """
 
     def __init__(
@@ -160,6 +182,7 @@ class UnifiProtectBackup:
         verify_ssl: bool,
         rclone_destination: str,
         retention: str,
+        ignore_cameras: List[str],
         verbose: int,
         port: int = 443,
     ):
@@ -177,10 +200,26 @@ class UnifiProtectBackup:
             retention (str): How long should event clips be backed up for. Format as per the
                             `--max-age` argument of `rclone`
                             (https://rclone.org/filtering/#max-age-don-t-transfer-any-file-older-than-this)
+            ignore_cameras (List[str]): List of camera IDs for which to not backup events
             verbose (int): How verbose to setup logging, see :func:`setup_logging` for details.
         """
         setup_logging(verbose)
 
+        logger.debug("Config:")
+        logger.debug(f"  {address=}")
+        logger.debug(f"  {port=}")
+        logger.debug(f"  {username=}")
+        if verbose < 5:
+            logger.debug("  password=REDACTED")
+        else:
+            logger.debug(f"  {password=}")
+
+        logger.debug(f"  {verify_ssl=}")
+        logger.debug(f"  {rclone_destination=}")
+        logger.debug(f"  {retention=}")
+        logger.debug(f"  {ignore_cameras=}")
+        logger.debug(f"  {verbose=}")
+
         self.rclone_destination = rclone_destination
         self.retention = retention
 
@@ -192,9 +231,9 @@ class UnifiProtectBackup:
             verify_ssl=verify_ssl,
             subscribed_models={ModelType.EVENT},
         )
+        self.ignore_cameras = ignore_cameras
         self._download_queue: asyncio.Queue = asyncio.Queue()
         self._unsub: Callable[[], None]
-        self._camera_names: Dict[str, str]
 
     async def start(self):
         """Bootstrap the backup process and kick off the main loop.
@@ -211,8 +250,10 @@ class UnifiProtectBackup:
         # Start the pyunifiprotect connection by calling `update`
         logger.info("Connecting to Unifi Protect...")
         await self._protect.update()
-        # Get a mapping of camera ids -> names
-        self._camera_names = {camera.id: camera.name for camera in self._protect.bootstrap.cameras.values()}
+        logger.info("Found cameras:")
+        for camera in self._protect.bootstrap.cameras.values():
+            logger.info(f" - {camera.id}: {camera.name}")
+
         # Subscribe to the websocket
         self._unsub = self._protect.subscribe_websocket(self._websocket_callback)
 
@@ -252,7 +293,7 @@ class UnifiProtectBackup:
         """Check if rclone is installed and the specified remote is configured.
 
         Raises:
-            RuntimeError: If rclone is not installed or it failed to list remotes
+            RcloneException: If rclone is not installed or it failed to list remotes
             ValueError: The given rclone destination is for a remote that is not configured
 
         """
@@ -271,7 +312,7 @@ class UnifiProtectBackup:
         logger.extra_debug(f"stdout:\n{stdout.decode()}")  # type: ignore
         logger.extra_debug(f"stderr:\n{stderr.decode()}")  # type: ignore
         if proc.returncode != 0:
-            raise RuntimeError(f"Failed to check rclone remotes: \n{stderr.decode()}")
+            raise RcloneException(stdout.decode(), stderr.decode(), proc.returncode)
 
         # Check if the destination is for a configured remote
         for line in stdout.splitlines():
@@ -295,6 +336,8 @@ class UnifiProtectBackup:
         assert isinstance(msg.new_obj, Event)
         if msg.action != WSAction.UPDATE:
             return
+        if msg.new_obj.camera_id in self.ignore_cameras:
+            return
         if msg.new_obj.end is None:
             return
         if msg.new_obj.type not in {EventType.MOTION, EventType.SMART_DETECT}:
@@ -316,27 +359,43 @@ class UnifiProtectBackup:
 
             logger.info(f"Backing up event: {event.id}")
             logger.debug(f"Remaining Queue: {self._download_queue.qsize()}")
-            logger.debug(f"  Camera: {self._camera_names[event.camera_id]}")
+            logger.debug(f"  Camera: {self._protect.bootstrap.cameras[event.camera_id].name}")
             logger.debug(f"  Type: {event.type}")
             logger.debug(f"  Start: {event.start.strftime('%Y-%m-%dT%H-%M-%S')}")
             logger.debug(f"  End: {event.end.strftime('%Y-%m-%dT%H-%M-%S')}")
             logger.debug(f"  Duration: {event.end-event.start}")
 
-            # TODO: Retry down/upload
             try:
                 # Download video
-                logger.debug("  Downloading video...")
-                video = await self._protect.get_camera_video(event.camera_id, event.start, event.end)
-            except Exception as e:
-                logger.warn("Failed to download video")
-                logger.exception(e)
-                continue
+                for x in range(5):
+                    try:
+                        logger.debug("  Downloading video...")
+                        video = await self._protect.get_camera_video(event.camera_id, event.start, event.end)
+                        assert isinstance(video, bytes)
+                        break
+                    except (AssertionError, aiohttp.client_exceptions.ClientPayloadError) as e:
+                        logger.warn(f"    Failed download attempt {x+1}, retying in 1s")
+                        logger.exception(e)
+                        await asyncio.sleep(1)
+                else:
+                    logger.warn(f"Download failed after 5 attempts, abandoning event {event.id}:")
+                    continue
 
-            try:
-                assert isinstance(video, bytes)
-                await self._upload_video(video, destination)
-            except RuntimeError:
-                continue
+                for x in range(5):
+                    try:
+                        await self._upload_video(video, destination)
+                        break
+                    except RcloneException as e:
+                        logger.warn(f"    Failed upload attempt {x+1}, retying in 1s")
+                        logger.exception(e)
+                        await asyncio.sleep(1)
+                else:
+                    logger.warn(f"Upload failed after 5 attempts, abandoning event {event.id}:")
+                    continue
+
+            except Exception as e:
+                logger.warn(f"Unexpected exception occured, abandoning event {event.id}:")
+                logger.exception(e)
 
     async def _upload_video(self, video: bytes, destination: pathlib.Path):
         """Upload video using rclone.
@@ -367,10 +426,7 @@ class UnifiProtectBackup:
             logger.extra_debug(f"stdout:\n{stdout.decode()}")  # type: ignore
             logger.extra_debug(f"stderr:\n{stderr.decode()}")  # type: ignore
         else:
-            logger.warn("Failed to download video")
-            logger.warn(f"stdout:\n{stdout.decode()}")
-            logger.warn(f"stderr:\n{stderr.decode()}")
-            raise RuntimeError(stderr.decode())
+            raise RcloneException(stdout.decode(), stderr.decode(), proc.returncode)
 
         logger.info("Backed up successfully!")
 
@@ -393,7 +449,7 @@ class UnifiProtectBackup:
         """
         path = pathlib.Path(self.rclone_destination)
         assert isinstance(event.camera_id, str)
-        path /= self._camera_names[event.camera_id]  # directory per camera
+        path /= self._protect.bootstrap.cameras[event.camera_id].name  # directory per camera
         path /= event.start.strftime("%Y-%m-%d")  # Directory per day
 
         file_name = f"{event.start.strftime('%Y-%m-%dT%H-%M-%S')} {event.type}"
